from typing import Any


class Lookup(object):
    def __init__(self, data: dict):
        self._data = data
        self.name = '45'

    def __getattr__(self, key: str):
        if key == '_': return {'dollar': '$'}
        value = self._data.get(key)
        if isinstance(value, dict):
            return Lookup(value)
        return value

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == '_data': super().__setattr__(__name, __value)
        else: self._data[__name] = __value


class Steps(object):
    def __init__(self, engine = 'sqlite'):
        self._first = None
        self._values = []
        self._misses = []
        self._counter = 0
        self.engine = engine
    
    @property
    def reset(self):
        self._counter = 0
        return self

    def next(self, count: int = 1) -> str:
        if self.engine == 'sqlite': return ', '.join(['?' for _ in range(count)])

        self._counter += 1
        return ', '.join([f'${i}' for i in range(self._counter, self._counter + (1 * count))])

    def EQUALS(self, key: str, datum: any):
        # necessary as sqlite uses 0 a falsy value for false
        if(datum == 0 or datum):
            if(self.dirty):
                sqls = f'AND {key} = {self.next()}'
            else:
                sqls = f'WHERE {key} = {self.next()}'
            self._values.append(datum)
            return sqls
        return ''

    def LIKE(self, key: str, datum: any):
        if(datum):
            if(self.dirty):
                sqls = f'AND {key} LIKE {self.next()}'
            else:
                sqls = f'WHERE {key} LIKE {self.next()}'
            self._values.append(f'%{datum}%')
            return sqls
        return ''

    @property
    def dirty(self) -> bool:
        return bool(self._values)

    @property
    def values(self) -> list:
        return self._values



from os import environ
from sqlite3 import Connection

from bcrypt import gensalt, hashpw
from heaven import Application, Response


async def cors(req, res: Response, ctx):
    allowed: str = req.headers.get('referer') or req.headers.get('X-Hosted') or ''
    hx_req_headers = 'HX-Boosted, HX-Current-URL, HX-History-Restore-Request, HX-Prompt, HX-Request, HX-Target, HX-Trigger-Name, HX-Trigger'
    hx_res_headers = 'HX-Location, HX-Push-Url, HX-Redirect, HX-Refresh, HX-Replace-Url, HX-Reswap, HX-Retarget, HX-Reselect, HX-Trigger, HX-Trigger-After-Settle, HX-Trigger-After-Swap'
    res.headers = 'Access-Control-Allow-Origin', allowed.strip('/')
    res.headers = 'Access-Control-Allow-Credentials', 'true'
    res.headers = 'Access-Control-Allow-Headers', f'Accept, Content-Type, Content-Disposition, Authorization, Authentication, Vary, Date, Accept-Encoding, X-CSRF-Token, X-Hint, X-Hosted, Set-Cookie, X-Form-ID, {hx_req_headers}'
    res.headers = 'Access-Control-Allow-Methods', 'GET, POST, PUT, PATCH, DELETE, OPTIONS'
    res.headers = 'Access-Control-Expose-Headers', f'X-Hint, X-Hosted, X-Other, Set-Cookie, X-Form-ID, HX-History-Restore-Request, {hx_res_headers}'


async def upsudo(app: Application) -> str:
    db: Connection = app.peek('db')
    username = environ.get('AMEBO_USERNAME')
    pwd = environ.get('AMEBO_PASSWORD')
    password = hashpw(pwd.encode(), gensalt())
    print("Administrator Password: ", pwd)

    # credentials table dropped every time so this is possible, what of producer credentials created after app start?
    # this has the potential to overwrite admin credentials if updated by admin after start? So document
    # to admins the need to set password from environmental variable and that that will always override and reset
    # everything else set afterwards
    try:
        if(app._.engine.startswith('postgres')):
            await db.execute(f'''
                INSERT INTO _amebo_.credentials(username, password) VALUES($1, $2)
                    ON CONFLICT(username) DO UPDATE SET password = EXCLUDED.password;
            ''', username, password.decode())
        else: db.execute('INSERT INTO credentials VALUES(?, ?);', (username, password.decode()))
    except Exception as exc:
        print("SUDO Credentials not created....")
        print('*' * 100, f': {exc}')


def upsecret(app: Application):
    if not environ.get('AMEBO_SECRET'):
        print('Deterministic dev secret key is: ', app.CONFIG('AMEBO_SECRET'))



from http import HTTPStatus
from sqlite3 import Connection

from fastjsonschema import JsonSchemaException, compile
from heaven import Context, Request, Response
from orjson import dumps, loads

from amebo.constants.literals import DB, MAX_PAGINATION
from amebo.decorators.formatters import jsonify
from amebo.decorators.providers import contextualize, expects
from amebo.models.events import Events
from amebo.utils.helpers import get_pagination, get_timeline
from amebo.utils.structs import Steps


@jsonify
@contextualize
async def tabulate(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    page, pagination = get_pagination(req)

    params = ['id', 'action', 'deduper', 'payload', 'timeline']
    _id, _action,  _deduper, _payload, _timeline = [req.queries.get(p) for p in params]

    steps = Steps(req.app._.engine)
    executor = ctx.executor

    sqls = f'''SELECT
            event, action, payload, deduper, timestamped, COUNT(*) AS results
        FROM {executor.schema}events
            {steps.EQUALS('event', _id)}
            {steps.LIKE('action', _action)}
            {steps.LIKE('payload', _payload)}
            {steps.EQUALS('deduper', _deduper)}
            {get_timeline(_timeline, steps)}
        GROUP BY
            event, action, deduper, payload, timestamped
        LIMIT {pagination if pagination < MAX_PAGINATION else MAX_PAGINATION}
        OFFSET {(page - 1) * pagination};
    '''
    try: rows = await executor.fetch(2).execute(sqls, *steps.values)
    except Exception as exc:
        return res.out(HTTPStatus.BAD_REQUEST, [])

    res.status = HTTPStatus.OK
    res.body = [{
        'event': event,
        'action': action,
        'payload': loads(payload),
        'deduper': deduper,
        'timestamped': timestamped
    } for event, action, payload, deduper, timestamped, results in rows]


@jsonify
@expects(Events)
@contextualize
async def insert(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    event: Events = ctx.events

    steps = Steps(req.app._.engine)
    try:
        executor = ctx.executor
        sqls = f'''
            SELECT
                schemata, actions.application, app.secret
            FROM
                {executor.schema}actions
            JOIN {executor.schema}applications app ON app.application = actions.application
            WHERE
                action = {steps.next()}
        '''
        row = await executor.fetch(1).execute(sqls, event.action)
        if not row: return res.out(HTTPStatus.UNPROCESSABLE_ENTITY, {'error': 'Action can not be used to process any events'})

        schemata = loads(row[0])  # load the schemata from the db
        if isinstance(schemata, str): schemata = loads(schemata)
        schemas = req.app.peek('schematas')  # get the compiled schematas
        if not schemas.get(event.action): schemas[event.action] = compile(schemata)
        validation = schemas.get(event.action)
        validation(event.payload)

        table = f'{executor.schema}events'
        fields = ['action', 'payload', 'deduper', 'timestamped',]
        values = [event.action, dumps(event.payload).decode(), event.deduper, event.timestamped.isoformat()]

        sqls = f'''INSERT INTO {table} ({', '.join(fields)}) VALUES ({steps.reset.next(4)}) RETURNING rowid;'''
        eventid = await executor.fetch(1).execute(sqls, *values)

        sqls = f'''
            INSERT INTO
                {executor.schema}gists(event, subscription, completed, retries, timestamped)
            SELECT
                {eventid[0]}, subscription, 0, 0, '{event.timestamped.isoformat()}'
            FROM {executor.schema}subscriptions WHERE action = {steps.reset.next()}
        '''
        await executor.fetch(0).execute(sqls, event.action)
    except JsonSchemaException:
        return res.out(HTTPStatus.NOT_ACCEPTABLE, {'error': f'Event payload does not conform to {event.action} schema'})
    except ModuleNotFoundError as exc:
        return res.out(HTTPStatus.UPGRADE_REQUIRED, {'error': f'{exc}'})

    res.status = HTTPStatus.CREATED
    res.body = {
        'event': eventid[0],
        'payload': event.payload,
        'action': event.action,
        'deduper': event.deduper,
        'timestamped': event.timestamped
    }



from asyncio import gather, sleep
from http import HTTPStatus
from sqlite3 import Connection, Cursor

# installed libs
from heaven import Router
from httpx import AsyncClient
from orjson import loads

# src code
from amebo.constants.literals import DB
from amebo.decorators.providers import Executor


async def aproko(router: Router):
    executor = Executor(router)
    x = executor.schema
    accepters = []
    rejecters = []
    async def notify(endpoint: str, data: dict, secret: str, gist_id: int):
        headers = {
            'Content-Type': 'application/json',
            'X-PASS-Phrase': secret
        }

        try:
            client = AsyncClient()
            result = await client.post(endpoint, json=data, headers=headers)
            if result.status_code not in [HTTPStatus.ACCEPTED, HTTPStatus.OK]: rejecters.append(int(gist_id))
            else: accepters.append(int(gist_id))
        except Exception as exc:
            print('Exception occured@@@@@@@@@@@@@@@@@@@@@@@@@: ', exc, ' ', endpoint)
            rejecters.append(gist_id)
        finally:
            await client.aclose()

        try:
            rejections = str(tuple(rejecters)).replace(',)', ')')
            sqls = f'''
                UPDATE {x}gists SET retries = retries + 1 WHERE rowid IN {rejections};
            '''
            if rejecters: await executor.fetch(0).execute(sqls)
        except Exception as exc: print('Could not negate in notify: ', exc)

        try:
            acceptances = str(tuple(accepters)).replace(',)', ')')
            sqls = f'''
                UPDATE {x}gists SET completed = 1, retries = retries + 1 WHERE rowid IN {acceptances};
            '''
            if accepters: await executor.fetch(0).execute(sqls)
        except Exception as exc: print('Could not update in notify: ', exc)

    async def traverse():
        try:
            gists = await executor.fetch(2).execute(f'''
                SELECT
                    s.handler AS endpoint, e.payload, a.secret, g.rowid as gid
                FROM {x}gists AS g JOIN {x}events e ON
                    g.event = e.event
                JOIN {x}subscriptions s ON
                    s.subscription = g.subscription
                JOIN {x}actions x ON
                    e.action = x.action
                JOIN {x}applications a ON
                    s.application = a.application
                WHERE g.completed <> 1
                AND g.retries < s.max_retries
                ORDER BY g.event LIMIT {router.CONFIG('envelope_size')};
            ''')

            if len(gists) < router.CONFIG('rest_when'): await sleep(router.CONFIG('idles'))
            await gather(*[notify(endpoint, loads(payload), secret, gid) for endpoint, payload, secret, gid in gists])
        except Exception as exc: print('Exception occured: ', exc)
    await traverse()
    return True


def cli():
    pass



from curses.ascii import isalnum
from datetime import datetime
from enum import Enum
from json import loads
from typing import Optional, Union

from pydantic import AnyHttpUrl
from pydantic import BaseModel as Model
from pydantic import Field, field_validator


class Credential(Model):
    username: str
    password: str
    scheme: str


class Application(Model):
    application: str = Field(min_length=1)
    address: AnyHttpUrl
    secret: str = Field(min_length=16)
    timestamped: datetime = Field(default_factory=datetime.now)

    @classmethod
    @field_validator('application')
    def no_spaces(cls, val: str) -> str:
        if val.count(' '): raise ValueError('Application name can not contain spaces')
        if not val.isalnum(): raise ValueError('Application name can only contain alphabets and numbers')
        return val

    @property
    def password(self):
        return '***'


class Location(Model):
    location: AnyHttpUrl
    secret: str


class Token(Model):
    secret: str
    microservice: str



from functools import wraps


def emits(name: str, properties: dict = None):
    def decorator(decorated: any):
        @wraps(decorated)
        def delegate(*args, **kwargs):
            return decorated(*args, **kwargs)
    return decorator




from functools import wraps
from http import HTTPStatus
from inspect import iscoroutinefunction
from sqlite3 import Connection

from asyncpg import Pool
from fastjsonschema import compile
from heaven import App, Context, Request, Response
from orjson import dumps, loads
from pydantic import BaseModel

from amebo.utils.structs import Lookup
from amebo.constants.literals import DB


class Executor(object):
    def __init__(self, app: App):
        self.app = app
        self.engine = self.app._.engine
        self.db = self.app._.db
        self._fetching = 0

    @property
    def schema(self):
        if self.engine == 'sqlite': return ''
        return '_amebo_.'

    def fetch(self, val: int):
        self._fetching = val
        return self

    async def _pg(self, query: str, *args):
        async with self.db.acquire() as conn:
            async with conn.transaction():
                if self._fetching == 1: return await conn.fetchrow(query, *args)
                if self._fetching > 1: return await conn.fetch(query, *args)
                else: return await conn.execute(query, *args)

    async def _sqlite(self, query: str, *args: tuple):
        cursor = self.db.cursor()
        try:
            if self._fetching == 1: return cursor.execute(query, args).fetchone()
            if self._fetching > 1: return cursor.execute(query, args).fetchall()
            else: return cursor.execute(query, args); self.db.commit()
        except Exception as exc: raise exc
        finally: cursor.close()

    def esc(self, count: int):
        return '?' if self.engine == 'sqlite' else f'${count}'

    @property
    def execute(self, *args, **kwargs):
        async def acaller(*args, **kwargs):
            if self.engine == 'sqlite': return await self._sqlite(*args, **kwargs)
            return await self._pg(*args, **kwargs)
        return acaller


def cacheschema(func):
    @wraps(func)
    async def delegate(req: Request, res: Response, ctx: Context):
        if iscoroutinefunction(func): await func(req, res, ctx)
        else: func(req, res, ctx)
        if req.app._.schematas.get(ctx.schemata_name): pass
        else: req.app._.schematas[ctx.schemata_name] = compile(ctx.schemata)
    return delegate


def contextualize(func):
    @wraps(func)
    async def delegate(req: Request, res: Response, ctx: Context):
        ctx.keep('executor', Executor(req.app))
        if iscoroutinefunction(func): await func(req, res, ctx)
        else: func(req, res, ctx)
    return delegate


def expects(Model: BaseModel):
    def wrapper(func):
        async def delegate(req: Request, res: Response, ctx: Context):
            appjson, ctype = 'application/json', 'content-type'
            res.headers = ctype, appjson
            if(req.headers.get(ctype) != appjson):
                res.status = HTTPStatus.IM_A_TEAPOT
                res.body = {'error': 'your request does not know json kung-fu'}
                return

            try:
                body = loads(req.body)
                model = Model(**body)
                mname = model.__class__.__name__.lower()
                ctx.keep(mname, model)
            except Exception as exc:
                error = exc.errors()[0]
                res.status = HTTPStatus.BAD_REQUEST
                print('-' * 20, error)
                res.body = {'error': f"{error.get('loc')[0]} - {error.get('msg')}"}
                print('-' * 10, res.body)
                return

            if iscoroutinefunction(func): await func(req, res, ctx)
            else: func(req, res, ctx)
        return delegate
    return wrapper


def inspects(schema: dict):
    validation = compile(schema)  # validictory?
    def decorator(func):
        @wraps(func)
        async def delegate(req: Request, res: Response, ctx: Context):
            try: payload = validation(loads(req.body))
            except: return res.out(HTTPStatus.NOT_ACCEPTABLE, 'Invalid data detected')
            ctx.keep('data', Lookup({key: value for key, value in payload.items() if value}))
            return await func(req, res, ctx)
        return delegate
    return decorator



pgscript = '''
CREATE SCHEMA IF NOT EXISTS _amebo_;

SET search_path TO _amebo_;
    CREATE TABLE IF NOT EXISTS _amebo_.credentials(
        rowid integer unique generated always as identity,
        username text primary key,
        password text NOT NULL
    );

    CREATE TABLE IF NOT EXISTS _amebo_.applications (
        rowid integer unique generated always as identity,
        application text primary key,
        address text NOT NULL,
        secret text NOT NULL,
        timestamped text NOT NULL
    );

    CREATE TABLE IF NOT EXISTS _amebo_.actions (
        rowid integer unique generated always as identity,
        action text primary key,
        application text NOT NULL REFERENCES applications(application),
        schemata text NOT NULL,
        timestamped text NOT NULL
    );

    CREATE TABLE IF NOT EXISTS _amebo_.events (
        rowid integer unique generated always as identity,
        event integer primary key generated always as identity,
        action text NOT NULL references actions(action),
        deduper text NOT NULL,
        payload text NOT NULL,
        timestamped text NOT NULL,

        UNIQUE(deduper, payload)
    );

    CREATE TABLE IF NOT EXISTS _amebo_.subscriptions (
        rowid integer unique generated always as identity,
        subscription integer primary key generated always as identity,  -- autogenerated for ease
        application text NOT NULL references applications(application),  -- subscribing app i.e. producer
        action text NOT NULL references actions(action),
        max_retries integer not null default 3,
        cron_interval text NOT NULL,
        handler text NOT NULL,
        description text,
        timestamped text NOT NULL,

        UNIQUE(application, action, handler)
    );

    CREATE TABLE IF NOT EXISTS _amebo_.gists (
        rowid integer primary key generated always as identity,
        event integer references events(event),
        subscription integer references subscriptions(subscription),
        completed integer NOT NULL,
        retries integer NOT NULL,
        timestamped text NOT NULL,

        UNIQUE(event, subscription)
    );

SET search_path TO public;
'''



from http import HTTPStatus
from inspect import iscoroutinefunction
from sqlite3 import Connection, Cursor, IntegrityError

# installed libs
from bcrypt import checkpw
from heaven import Context, Request, Response

# src code
from amebo.constants.literals import DB, AMEBO_SECRET, MAX_PAGINATION
from amebo.decorators.formatters import jsonify
from amebo.decorators.providers import contextualize, expects
from amebo.utils.helpers import get_pagination, get_timeline, tokenize

from amebo.models.applications import Credential, Location, Application, Token
from amebo.utils.structs import Steps


@jsonify
@expects(Credential)
@contextualize
async def authenticate(req: Request, res: Response, ctx: Context):
    def unauthorized():
        res.status = HTTPStatus.UNAUTHORIZED
        res.body = {'error': 'could not authenticate microservice'}

    db: Connection = req.app.peek(DB)
    credential: Credential = ctx.credential
    executor = ctx.executor

    table = 'credentials'
    username_field = 'username'
    password_field = 'password'
    if credential.scheme == 'token':
        table = 'applications'
        username_field = 'application'
        password_field = 'secret'

    try:
        SQL = f'''
            SELECT {username_field}, {password_field} FROM {executor.schema}{table}
                WHERE {username_field} = {executor.esc(1)}
        '''
        row = await executor.fetch(1).execute(SQL, (credential.username))
    except Exception as exc:
        return unauthorized()

    if not row: return unauthorized()
    username, password = row
    if not checkpw(credential.password.encode(), password.encode()): return unauthorized()
    # no feedback is provided if secret key mismatches i.e. continue indicates just that

    token = tokenize({
        'scheme': credential.scheme,
        'username': username,
    }, req.app.CONFIG('AMEBO_SECRET'))
    res.headers = 'Set-Cookie', f'Authentication={token}; Path=/; HttpOnly; Max-Age={60*10}; SameSite=Strict; Secure'
    res.status = HTTPStatus.ACCEPTED
    res.body = {'token': token}


@jsonify
@contextualize
async def tabulate(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    page, pagination = get_pagination(req)
    params = ['application', 'address', 'timeline']
    _application, _address, _timeline = [req.queries.get(p) for p in params]
    steps = Steps(req.app._.engine)

    executor = ctx.executor

    sqls = f'''SELECT application, address, timestamped
        FROM {executor.schema}applications
            {steps.LIKE('application', _application)}
            {steps.LIKE('address', _address)}
            {get_timeline(_timeline, steps)}
        LIMIT {pagination if pagination < MAX_PAGINATION else MAX_PAGINATION}
        OFFSET {(page - 1) * pagination};
    '''
    try: rows = await executor.fetch(2).execute(sqls, *steps.values)
    except Exception as exc:
        res.status = HTTPStatus.BAD_REQUEST
        res.body = {'error': f'{exc}'}
        return

    res.status = HTTPStatus.OK
    res.body = [{
        'application': application,
        'address': address,
        'secret': '****************',
        'timestamped': timestamped
    } for application, address, timestamped in rows]


@jsonify
@expects(Application)
@contextualize
async def insert(req: Request, res: Response, ctx: Context):
    application: Application = ctx.application
    steps = Steps(req.app._.engine)
    try:
        executor = ctx.executor
        values = (str(application.application), str(application.address), application.secret, application.timestamped.isoformat())
        sqls = f'''INSERT INTO {executor.schema}applications(application, address, secret, timestamped) VALUES ({steps.next(4)})'''
        print(sqls)
        await executor.execute(sqls, *values)
    except IntegrityError as exc:
        res.status = HTTPStatus.CONFLICT
        res.body = {'error': f'{exc}'}
        return
    except Exception as exc:
        res.status = HTTPStatus.NOT_ACCEPTABLE
        res.body = {'error': f'{exc}'}
        return

    res.status = HTTPStatus.CREATED
    res.body = {
        'name': application.application,
        'address': str(application.address),
        'secret': application.secret,
        'timestamped': application.timestamped
    }


@jsonify
@expects(Location)
@contextualize
def update(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    application = req.params.get('id')
    location: Location = ctx.location

    executor = ctx.executor()
    try:
        cursor = db.cursor()
        cursor.execute(f'''
            UPDATE applications SET address = {executor.next()}
                WHERE application = {executor.next()} AND secret = {executor.next()}
        ''', (location.location, application, location.secret))
    except Exception as exc:
        res.status = HTTPStatus.BAD_REQUEST
        res.body = {'error': 'could not update application'}
        return
    finally:
        cursor.close()

    # no feedback if provided if secret key mismatches i.e. continue indicates just that
    res.status = HTTPStatus.ACCEPTED
    res.body = None


@jsonify
@expects(Token)
def tokens(req: Request, res: Response, ctx: Context):
    # TODO: set cookie and return body at same time for browser and api
    token: Token = ctx.token
    db: Connection = req.app.peek(DB)

    try:
        cursor = db.cursor()
        cursor.execute(f'''
            SELECT application, secret FROM applications WHERE application = ?
        ''', (token.application)).fetchall()
    except Exception as exc:
        res.status = HTTPStatus.BAD_REQUEST
        res.body = {'error': 'could not update application'}
        return
    finally:
        if cursor: cursor.close()

    res.status = HTTPStatus.ACCEPTED
    res.headers = 'set-cookie', '#cookie body here'
    res.body = {'token': token}



from http import HTTPStatus
from inspect import iscoroutinefunction

from heaven import Context, Request, Response
from jwt import decode, encode

from amebo.constants.literals import AMEBO_SECRET


async def authenticate(req: Request, res: Response, ctx: Context):
    sk = req.app.CONFIG(AMEBO_SECRET)
    def leave():
        res.headers = 'Location', '/'
        return res.out(HTTPStatus.TEMPORARY_REDIRECT, None)

    # TODO: remove secret from ui and use withCredentials as we use the cookie to get it from the tokens cache
    #TODO: Change checking for secret in handlers to checking for producer name as authorization and identification is done here
    authentication = req.cookies.get('authentication')
    if not authentication: return leave()

    try: metadata = decode(authentication, sk, algorithms='HS256')
    except: return leave()

    # keep metadata for use later in pages if required
    ctx.keep('metadata', metadata)


def authorization(func):
    async def delegate(req: Request, res: Response, ctx: Context):
        sk = req.app.CONFIG(AMEBO_SECRET)
        authorization = req.headers.get('authorization')
        authentication = req.cookies.get('Authentication')

        if not authorization or authentication: return res.out(HTTPStatus.UNAVAILABLE_FOR_LEGAL_REASONS, {'error': 'No credentials provided'})
        if iscoroutinefunction(func): return await func(req, res, ctx)

        try:
            if authorization: _, token = authorization.split(' ', 1)
            else: token = authentication
        except: return res.out(HTTPStatus.BAD_REQUEST, {'error': 'Could not process your credentials'})

        try: metadata = decode(token, sk, algorithms='HS256')
        except: return res.out(HTTPStatus.UNAUTHORIZED, {'error': 'Invalid credentials detected'})

        # user producer name to use for token lookup from global app state
        producer = metadata.get('producer')
        token = req.app._.tokens._data.get(producer)
        if not token: return res.out(HTTPStatus.UNAUTHORIZED, {'error': 'Application credentials cache cleared at app startup'})

        ctx.keep('token', token)
        return func(req, res, ctx)
    return delegate


def protected(func):
    async def delegate(req: Request, res: Response, ctx: Context):
        sk = req.app.CONFIG(AMEBO_SECRET)
        def leave():
            res.headers = 'Location', '/'
            return res.out(HTTPStatus.TEMPORARY_REDIRECT, None)

        # TODO: remove secret from ui and use withCredentials as we use the cookie to get it from the tokens cache
        #TODO: Change checking for secret in handlers to checking for producer name as authorization and identification is done here
        authentication = req.cookies.get('Authentication')
        # import pdb; pdb.set_trace()
        if not authentication: return leave()

        try: metadata = decode(authentication, sk, algorithms='HS256')
        except: return leave()

        # keep metadata for use later in pages if required
        ctx.keep('metadata', metadata)

        # nothing else to see here move on to next delegate or handler
        if iscoroutinefunction(func): return await func(req, res, ctx)
        else: return func(req, res, ctx)
    return delegate



from inspect import iscoroutinefunction
from os import environ
from sqlite3 import Connection

from heaven import Application
from asyncpg import create_pool

from amebo.constants.literals import DB
from amebo.constants.scripts import initdbscript
from amebo.utils.structs import Lookup
from amebo.database.pg import pgscript


ENGINES = {
    'sqlite': lambda *args: Connection(*args),
    'postgres': lambda *args: create_pool(*args)
}


async def connect(app: Application):
    engine = app.CONFIG('engine').lower()
    app.keep('engine', engine)
    try:
        if engine.startswith('postgres'): db = await create_pool(environ.get('AMEBO_DSN'))
        else: db = Connection('amebo.db')
    except Exception as exc: print('connection middleware failed: {exc}')
    app.keep(DB, db)


async def disconnect(app: Application):
    try: db: Connection = app.peek(DB)
    except Exception as exc: print('Exception in closing db on shutdown: ', exc)
    else:
        if(iscoroutinefunction(db.close)): await db.close()
        else: db.close()


async def initialize(app: Application):
    """todo: enable switching db backend between redis, pg, sqlite"""
    if app._.engine.startswith('postgres'):
        await app.peek(DB).execute(pgscript)
    else:
        try:
            db: Connection = app.peek(DB)
            db.executescript(initdbscript)
        except Exception as exc:
            print('Exception in initdb hook: ', exc)


def cache(app: Application):
    app._.tokens = {}
    app._.schematas = {}



initdbscript = '''
BEGIN;
    DROP TABLE IF EXISTS credentials;

    CREATE TABLE IF NOT EXISTS credentials(
        username text primary key,
        password text NOT NULL
    );

    CREATE TABLE IF NOT EXISTS applications (
        application text primary key,
        address text NOT NULL,
        secret text NOT NULL,
        timestamped text NOT NULL
    );

    CREATE TABLE IF NOT EXISTS actions (
        action text primary key,
        application text NOT NULL REFERENCES applications(application),
        schemata text NOT NULL,
        timestamped text NOT NULL
    );

    CREATE TABLE IF NOT EXISTS events (
        event integer primary key NOT NULL,
        action text NOT NULL references actions(action),
        deduper text NOT NULL,
        payload text NOT NULL,
        timestamped text NOT NULL,

        UNIQUE(deduper, payload)
    );

    CREATE TABLE IF NOT EXISTS subscriptions (
        subscription integer primary key NOT NULL,  -- autogenerated for ease
        application text NOT NULL references applications(application),  -- subscribing app i.e. producer
        action text NOT NULL references actions(action),
        max_retries integer not null default 3,
        handler text NOT NULL,
        description text,
        timestamped text NOT NULL,

        UNIQUE(application, action, handler)
    );

    CREATE TABLE IF NOT EXISTS gists (
        event integer references events(event),
        subscription integer references subscriptions(subscription),
        completed integer NOT NULL,
        retries integer NOT NULL,
        timestamped text NOT NULL,

        UNIQUE(event, subscription)
    );
COMMIT;
'''

insertscript = '''
    INSERT INTO credentials VALUES(?, ?)
'''

tidydatabase = '''
    CREATE TRIGGER IF NOT EXISTS
        prune_gists
    AFTER INSERT ON
        gists
    WHEN
        count(*) > 1_000_000
    BEGIN
        DELETE FROM gists WHERE timestamped < '14 days ago'
    END
'''



from http import HTTPStatus
from sqlite3 import Connection, Cursor

# installed libs
from heaven import Context, Request, Response
from orjson import dumps, loads

from amebo.constants.literals import DB, MAX_PAGINATION
from amebo.decorators.formatters import jsonify
from amebo.decorators.providers import contextualize, expects
from amebo.decorators.providers import cacheschema
from amebo.models.actions import Action
from amebo.utils.helpers import get_pagination, get_timeline
from amebo.utils.structs import Steps


@jsonify
@contextualize
async def tabulate(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    page, pagination = get_pagination(req)
    params = ['id', 'action', 'application', 'schemata', 'timeline']
    _id, _action, _application, _schemata, _timeline = [req.queries.get(p) for p in params]

    steps = Steps(req.app._.engine)
    executor = ctx.executor

    sqls = f'''
        SELECT
            rowid, action, application, schemata, timestamped
        FROM
            {executor.schema}actions
            {steps.EQUALS('rowid', _id)}
            {steps.LIKE('action', _action)}
            {steps.LIKE('application', _application)}
            {steps.LIKE('schemata', _schemata)}
            {get_timeline(_timeline, steps)}
        LIMIT {pagination if pagination < MAX_PAGINATION else MAX_PAGINATION}
        OFFSET {(page - 1) * pagination};
    '''
    try: rows = await executor.fetch(2).execute(sqls, *steps.values)
    except Exception as exc:
        res.status = HTTPStatus.BAD_REQUEST
        res.body = {'error': f'{exc}'}
        return

    res.status = HTTPStatus.OK
    res.body = [{
        'id': id,
        'action': action,
        'application': application,
        'schemata': loads(schemata),
        'timestamped': timestamped
    } for id, action, application, schemata, timestamped in rows]


@jsonify
@expects(Action)
@contextualize
async def insert(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    action: Action = ctx.action
    
    steps = Steps(req.app._.engine)
    executor = ctx.executor

    table = 'actions'
    fields = ('action', 'application', 'schemata', 'timestamped',)
    values = (action.action, action.application, dumps(action.schemata).decode(), action.timestamped.isoformat())

    try:
        sqls = f'''select application, secret from {executor.schema}applications where application = {steps.next()}'''
        _application = await executor.fetch(1).execute(sqls, action.application)
        if not _application:
            raise ValueError(f'Application {action.application} not found')
        application, secret = _application
        if(secret != action.secret):
            return res.out(HTTPStatus.UNAUTHORIZED, {'error': f'Incorrect {application} secret detected'})
    except Exception as exc: return res.out(HTTPStatus.UNAUTHORIZED, {'error': f'{exc}'})
    
    try:
        sqls = f'''INSERT INTO {executor.schema}{table}({', '.join(fields)}) VALUES ({steps.reset.next(4)})'''
        print(sqls)
        await executor.fetch(0).execute(sqls, *values)
    except Exception as exc: return res.out(HTTPStatus.UPGRADE_REQUIRED, {'error': f'{exc}'})

    ctx.keep('schemata', action.schemata)
    res.status = HTTPStatus.CREATED
    res.body = action.model_dump()




from datetime import datetime, timezone
from http import HTTPStatus
from sqlite3 import Connection, Cursor

from heaven import Context, Request, Response

from amebo.constants.literals import DB, MAX_PAGINATION
from amebo.decorators.formatters import jsonify
from amebo.decorators.providers import contextualize, expects
from amebo.models.subscriptions import Subscriptions
from amebo.utils.helpers import get_pagination, get_timeline
from amebo.utils.structs import Steps


@jsonify
@contextualize
async def tabulate(req: Request, res, ctx: Context):
    db: Connection = req.app.peek(DB)
    page, pagination = get_pagination(req)
    qstrings = ['id', 'application', 'matchrule', 'action', 'endpoint', 'description', 'timeline']
    [
        _id, _application, _matchrule,
        _action, _handler, _description, _timeline] = [req.queries.get(q) for q in qstrings]

    steps = Steps(req.app._.engine)
    executor = ctx.executor
    sqls = f'''
        SELECT
            subscription, action, application, handler, description, timestamped
        FROM {executor.schema}subscriptions
            {steps.EQUALS('subscription', _id)}
            {steps.LIKE('application', _application)}
            {steps.LIKE('action', _action)}
            {steps.LIKE('handler', _handler)}
            {steps.LIKE('description', _description)}
            {get_timeline(_timeline, steps)}
        LIMIT {pagination if pagination < MAX_PAGINATION else MAX_PAGINATION}
        OFFSET {(page - 1) * pagination};
    '''
    try: rows = await executor.fetch(2).execute(sqls, *steps.values)
    except Exception as exc:
        res.status = HTTPStatus.BAD_REQUEST
        res.body = None
        return

    res.status = HTTPStatus.OK
    res.body = [{
        'subscription': subscription,
        'action': action,
        'application': application,
        'endpoint': handler,
        'description': description,
        'timestamped': timestamped
    } for subscription, action, application, handler, description, timestamped in rows]


@jsonify
@expects(Subscriptions)
@contextualize
async def insert(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    subscriptions: Subscriptions = ctx.subscriptions
    
    steps = Steps(req.app._.engine)
    executor = ctx.executor

    try:
        sqls = f'SELECT address FROM {executor.schema}applications WHERE application = {steps.next()} AND secret = {steps.next()}'
        rows = await executor.fetch(1).execute(sqls, *(subscriptions.application, subscriptions.secret))
    except Exception as exc:
        return res.out(HTTPStatus.BAD_REQUEST, {'error': f'Invalid data submmitted {exc}'})

    if not rows: return res.out(HTTPStatus.NOT_ACCEPTABLE, {'error': 'Subscription request rejected'})
    try:
        row = rows[0]
        host = row.strip('/')
    except: return res.out(HTTPStatus.UNPROCESSABLE_ENTITY, 'Can not process the event with information provided')
    address = f'{host}{subscriptions.handler}'

    fields = ('application', 'action', 'max_retries', 'handler', 'timestamped',)
    values = (
        subscriptions.application,  # subscribing application
        subscriptions.action,
        subscriptions.max_retries,
        address,
        datetime.now(tz=timezone.utc).isoformat()
    )

    try:
        sqls = f'''INSERT INTO {executor.schema}subscriptions ({', '.join(fields)}) VALUES ({steps.reset.next(5)}) RETURNING rowid;'''
        subscriptionid = await executor.execute(sqls, *values)
    except Exception as exc: return res.out(HTTPStatus.UPGRADE_REQUIRED, {'error': f'{exc}'})

    res.status = HTTPStatus.CREATED
    subscriptions.subscription = subscriptionid[0]
    
    res.body = subscriptions.model_dump()



from http import HTTPStatus
from sqlite3 import Connection, Cursor, IntegrityError

from heaven import Context, Request, Response
from httpx import AsyncClient
from orjson import loads

from amebo.decorators.formatters import jsonify
from amebo.decorators.security import protected
from amebo.decorators.providers import contextualize
from amebo.constants.literals import DB, MAX_PAGINATION
from amebo.utils.helpers import get_pagination, get_timeline
from amebo.utils.structs import Steps


@jsonify
@contextualize
async def tabulate(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    page, pagination = get_pagination(req)
    params = ['origin', 'destination', 'gist', 'event', 'completed', 'timeline']
    _origin, _destination, _gist, _event, _completed, _timeline = [req.queries.get(p) for p in params]
    _completed = _completed or 'all'

    completed = {'all': '', 'true': 1, 'false': 0}.get(_completed.lower())

    steps = Steps(req.app._.engine)
    executor = ctx.executor
    sqls = f'''
        SELECT
            g.rowid as gist,
            e.action as action,
            case when
                g.completed <> 0
            then
                'True'
            else
                'False'
            end as
                completed,
            x.application as publisher,
            s.application as subscriber,
            g.timestamped
        FROM {executor.schema}gists AS g JOIN {executor.schema}events e ON
            g.event = e.event
        JOIN {executor.schema}subscriptions s ON
            g.subscription = s.subscription
        JOIN {executor.schema}actions x ON
            s.action = x.action
        {steps.EQUALS('g.rowid', _gist)}
        {steps.LIKE('e.producer', _origin)}
        {steps.EQUALS('g.completed', completed)}
        {steps.LIKE('a.event', _event)}
        {steps.LIKE('p.name', _destination)}
        {get_timeline(_timeline, steps, column='g.timestamped')}
        ORDER BY g.rowid
        LIMIT {pagination if pagination < MAX_PAGINATION else MAX_PAGINATION}
        OFFSET {(page - 1) * pagination};
    '''
    try:
        rows = await executor.fetch(2).execute(sqls, *steps.values)
    except Exception as exc:
        print('exc is: ', exc)
        res.status = HTTPStatus.BAD_REQUEST
        res.body = {'error': f'{exc}'}
        return

    res.status = HTTPStatus.OK
    res.body = [{
        'gist': gist,
        'action': action,
        'completed': completed,
        'publisher': publisher,
        'subscriber': subscriber,
        'timestamped': timestamped
    } for gist, action, completed, publisher, subscriber, timestamped in rows]


@jsonify
@contextualize
async def replay(req: Request, res: Response, ctx: Context):
    db: Connection = req.app.peek(DB)
    id = req.params.get('id')
    
    steps = Steps(req.app._.engine)
    executor = ctx.executor

    try:
        gist = await executor.fetch(1).execute(f'''
            SELECT
                s.handler AS endpoint, e.payload, a.secret, g.rowid as gid
            FROM gists AS g JOIN subscriptions s ON
                g.subscription = s.subscription
            JOIN events e ON
                g.event = e.event
            JOIN applications a ON
                s.application = a.application
            WHERE g.rowid = {steps.next()};
        ''', (id,))
    except Exception as exc:
        res.status = HTTPStatus.BAD_REQUEST
        res.body = {'error': f'{exc}'}
        return

    if not gist: return res.out(HTTPStatus.NOT_FOUND, {'error': 'Gist not found'})

    try:
        sender = AsyncClient()

        endpoint, payload, secret, gid = gist
        headers = {'content-type': 'application/json', 'x-pass-phrase': secret}

        response = await sender.post(endpoint, json=loads(payload), headers=headers)
        if response.status_code not in [HTTPStatus.ACCEPTED, HTTPStatus.OK]:
            raise ConnectionRefusedError('Endpoint maybe offline, failed to handle gist')
    except ConnectionRefusedError as exc:
        res.status = HTTPStatus.SERVICE_UNAVAILABLE
        try: proxied = response.json()
        except: proxied = None
        res.body = {'gist': gid, 'proxied': proxied, 'error': f'{exc}'}
        return
    except Exception as exc:
        res.status = HTTPStatus.BAD_GATEWAY
        res.body = {'error': f'{exc}'}
        return
    finally:
        await sender.aclose()

    res.satus = HTTPStatus.ACCEPTED
    try: proxied = response.json()
    except: proxied = None
    res.body = {'gist': gid, 'proxied': proxied}
    return



from os import environ
from uuid import uuid4

# installed libs
from heaven import Application
from heaven.constants import STARTUP, SHUTDOWN

# src code
from amebo.aproko import aproko
from amebo.constants.literals import AMEBO_SECRET
from amebo.utils.helpers import deterministic_uuid


router = Application({
    'engine': environ.get('AMEBO_DSN') or 'sqlite',
    'envelope_size': int(environ.get('AMEBO_ENVELOPE') or 256),  # how many tasks to fetch at once for processing
    'idles': 5,  # sleep for 5 seconds
    'rest_when': 0,  # reduce frequency of daemons when tasks less than 5
    AMEBO_SECRET: environ.get('AMEBO_SECRET') or deterministic_uuid()
})


# setup daemons and app engine helpers
router.daemons = aproko
router.ASSETS('public', relative_to=__file__)
router.TEMPLATES('templates', relative_to=__file__)


# set up hooks
router.ON(STARTUP, 'amebo.middlewares.database.connect')
router.ON(STARTUP, 'amebo.middlewares.database.cache')
router.ON(SHUTDOWN, 'amebo.middlewares.database.disconnect')
router.ON(STARTUP, 'amebo.middlewares.database.initialize')
router.ON(STARTUP, 'amebo.middlewares.security.upsudo')
router.ON(STARTUP, 'amebo.middlewares.security.upsecret')


# hooks
router.BEFORE('/*', 'amebo.middlewares.security.cors')


# authenticate first
router.POST('/v8/tokens', 'amebo.controllers.applications.authenticate')


# web ui- views/pages/screens
router.GET('/', 'amebo.controllers.xui.login')
router.GET('/p/:page', 'amebo.controllers.xui.pages')


# api
router.GET('/v1/actions', 'amebo.controllers.actions.tabulate')
router.GET('/v1/events', 'amebo.controllers.events.tabulate')
router.GET('/v1/applications', 'amebo.controllers.applications.tabulate')
router.GET('/v1/subscriptions', 'amebo.controllers.subscriptions.tabulate')
router.GET('/v1/gists', 'amebo.controllers.gists.tabulate')
router.POST('/v1/tokens', 'amebo.controllers.applications.authenticate')
router.POST('/v1/actions', 'amebo.controllers.actions.insert')
router.POST('/v1/events', 'amebo.controllers.events.insert')
router.POST('/v1/applications', 'amebo.controllers.applications.insert')
router.POST('/v1/subscriptions', 'amebo.controllers.subscriptions.insert')
router.POST('/v1/regists/:id', 'amebo.controllers.gists.replay')
router.PUT('/v1/applications/:id', 'amebo.controllers.applications.update')

# maybe add a route to clear cache of compiled schemas ?


# comment me out in production
# router.POST('/h1/identity-created', amebo_sleeper)



from datetime import datetime
from json import loads
from typing import Optional, Union

from pydantic import AnyHttpUrl
from pydantic import BaseModel as Model
from pydantic import Field, field_validator

from amebo.models.actions import Action


class Subscriptions(Model):
    subscription: Optional[int] = None
    application: str
    action: str
    handler: str
    secret: str
    max_retries: int = Field(le=10_000, ge=1)
    timestamped: datetime = Field(default_factory=datetime.now)

    @field_validator('handler')
    @classmethod
    def _handler(cls, val: str):
        if not val.startswith('/'): raise ValueError('Subscription handlers must start with a leading `/`')
        return val


class Gists(Model):
    attempts: int = 0
    action: Action
    subscription: Subscriptions
    status: int
